def beautify_num(x):
	"""
	:type x: float or int
	:rtype: str
	"""
	if abs(x) >= 100 or abs(x) <= 0.01:
		value_str = "{:.2e}".format(x)
	elif abs(x) >= 1:
		value_str = "{:.3f}".format(x)
	elif abs(x) >= 0.1:
		value_str = "{:.4f}".format(x)
	else:
		value_str = "{:.4f}".format(x)
	return value_str





from copy import deepcopy


def apply_on_dict_keys(dictionary, func, recursive=True):
	"""
		:type dictionary: dict
		:type func: callable
		:rtype: dict
		"""
	dictionary = deepcopy(dictionary)
	if isinstance(dictionary, dict):

		keys = list(dictionary.keys())
		for old_key in keys:

			new_key = func(old_key)
			old_value = dictionary.pop(old_key)
			if recursive:
				new_value = apply_on_dict_keys(dictionary=old_value, func=func, recursive=recursive)
			else:
				new_value = old_value
			dictionary[new_key] = new_value

		return dictionary

	elif isinstance(dictionary, list):
		if recursive:
			return [apply_on_dict_keys(dictionary=d, func=func, recursive=recursive) for d in dictionary]
		else:
			return dictionary

	else:
		return dictionary







from collections import Counter


def has_duplicates(l):
	"""
	:type l: list
	:rtype: bool
	"""
	if len(l) == 0:
		return False
	else:
		return Counter(l).most_common()[0][1] > 1


def get_duplicates(l):
	"""
	:type l: list
	:rtype: list
	"""
	return [k for k, v in Counter(l).items() if v > 1]





def merge_dictionaries(x, y):
    """
    :type x: dict
    :type y: dict
    :rtype: dict
    """
    z = x.copy()   # start with x's keys and values
    z.update(y)    # modifies z with y's keys and values & returns None
    return z






# The Dictionary class is the same as dict with additional methods


class Dictionary(dict):

	def copy(self):
		return Dictionary(super().copy())

	# the map method allows you to run a function on all values of the dictionary
	def map(self, func, inplace = False):
		if inplace:
			copy = self
		else:
			copy = self.copy()

		try:
			for key in copy:
				copy[key] = func(copy[key])
		except TypeError:
			return False

		if inplace==False:
			return copy

	# based on a solution on stackoverflow:
	# https://stackoverflow.com/questions/12229064/mapping-over-values-in-a-python-dictionary

	@staticmethod
	def from_lists(keys, values = None):
		if values is None:
			values = keys

		return Dictionary(zip(keys, values))

	@staticmethod
	def from_dict(x):
		return Dictionary.from_lists(keys = x.keys(), values = x.values())





from math import sin, cos, sqrt, atan2, radians
import geopy.distance
earth_radius = 6373  # km

def get_haversine_distance(latitude1, longitude1, latitude2, longitude2, unit = 'km'):
	lat1 = radians(latitude1)
	lon1 = radians(longitude1)
	lat2 = radians(latitude2)
	lon2 = radians(longitude2)

	dlon = lon2 - lon1
	dlat = lat2 - lat1

	a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
	c = 2 * atan2(sqrt(a), sqrt(1 - a))

	distance = earth_radius * c

	if unit.lower() != 'km':
		distance *= 0.621371

	return distance

def get_vincenty_distance(latitude1, longitude1, latitude2, longitude2, unit = 'km'):
	coords_1 = (latitude1, longitude1)
	coords_2 = (latitude2, longitude2)
	distance = geopy.distance.vincenty(coords_1, coords_2).km

	if unit.lower() != 'km':
		distance *= 0.621371

	return distance

def get_distance(latitude1, longitude1, latitude2, longitude2, unit = 'km', method = 'vincenty'):
	if method.lower() == 'vincenty':
		return get_vincenty_distance(
			latitude1 = latitude1, longitude1 = longitude1,
			latitude2 = latitude2, longitude2 = longitude2,
			unit = unit
		)
	else:
		return get_haversine_distance(
			latitude1=latitude1, longitude1=longitude1,
			latitude2=latitude2, longitude2=longitude2,
			unit=unit
		)





from datetime import datetime
import os
import errno

class TimeStamp:
	def __init__(self, date = datetime.today(), prefix = '', suffix = '', echo = True):
		self._date = date
		self._prefix = prefix
		self._suffix = suffix
		self._dir = None
		self._echo = echo

	def stamp(self, date=datetime.today(), format = "%Y-%m%d-%H%M"):
		the_stamp = self._prefix + self._date.strftime(format) + self._suffix
		if self._echo:
			print('stamp:', the_stamp)
		return the_stamp

	def make_dir(self):
		dir_name = self.stamp()

		try:
			os.makedirs(dir_name)
		except OSError as e:
			if e.errno != errno.EEXIST:
				raise
		if self._echo:
			print('make dir:', dir_name)
		self._dir = dir_name

	def get_path(self, file):
		the_path = self._dir + '/' + file
		if self._echo:
			print('file path:', the_path)
		return the_path
